;***************************************************************************
;
;                          CONDS.LSP
;
;  This file contains the condition-types the step needs to be able to test.
; There are essentially 8 kinds of tests that can be used. Many of them only 
; take one argument, but some take 2 arguments. The 8 tests are listed below,
; with the number and style of arguments each takes.
;
;     Class-name (args)                         short-form definition
;
;     word-cond (necs &sups)                   wordc
;     or-cond   (necs)                         orc
;     and-cond  (necs &sups)                   andc
;     not-cond  (necs)                         notc
;     max-cond  (necs)                         maxc
;     special-score-cond (&score &necs )       spscorec
;     history-cond (necs)                      historyc
;     lambda-cond  (necs)                      lambdac
;
; Legend:'necs' (necessary-conditions) represents arguments that must be 
;        recognized 
;           in order for the test to succeed 
;       'sups' (supporting conditions) means the test may succeed even if
;          the evaluated sups argument does not.
;       '&' represents optional arguments. For example, it is not necessary
;          to provide a set of sups for any test.
;       'score' is used only with special-score-cond. The other types of 
;          conditions each have their own mechanism for scoring. The
;          special-score-cond allows the knowledge engineer to override this 
;          score.
;
; Note: the short-form definitions with one argument do not need keywords;
;       the ones with two arguments take :sc, :nec or :sup, as applicable.
;
; Note: the lambda inside the lambda-cond should expect one argument of the 
;       form (:restaurant <the restaurant>), where :restaurant is a keyword
;       and <the restaurant> is an object.
;
;***************************************************************************


;**********************************************************************
;*                                                                    *
;*                  STEP-CONDITION CLASS                              *
;*                                                                    *
;*  This is just an abstract class. Its do-test 'ands' sub-conditions *
;**********************************************************************


(defclass step-condition ()
  ((necessary-conditions
    :accessor necessary-conditions
    :initform '()
    :initarg :necessary-conditions)
   (supporting-conditions
    :accessor supporting-conditions
    :initform '()
    :initarg :supporting-conditions)))

(defmethod do-test ((self step-condition) restaurant step)
     :documentation
         "Test the conditions on the step. Any results from the test are 
          built up in the result-variable. Results include the score for how 
          much was recognized, as well as any bindings that were generated as 
          a result of the test. The generic step-condition 'ands' the 
          necessary conditions and sums the supporting conditions. 
          Note that the conditions
          (necessary and supporting) are a list of functions of any length" 
      
	 (let ((results (make-instance 'results :success t 
                        :remaining-input (input (waiter restaurant))))
            (more-results))
         (loop for c in (necessary-conditions self) do
               (setf more-results (apply #'do-test (list c restaurant step)))
               (cond ((success more-results) 
                      (cond ((listp (score more-results))
                             (add-to-score results (score more-results)))
                            (t 
                               (add-score results (score more-results))))
                 ;     (format t "~% bindings of more-results: ~a" 
                 ;        (bindings more-results))
                      (add-bindings results (bindings more-results))
                 ;     (format t "~%new bindings: ~a" (bindings results))
                      (remove-bindings-from-remaining-input results 
						    (bindings more-results))
                 ;     (format t "~%Remaining input (from do-test): ~a" 
                  ;                            (remaining-input results))    
                      )
                     (t (setf (success results) '()))))
       (when debugp (format t "~%  results in do-test: ~a" results))
       (when (success results)
         (loop for c in (supporting-conditions self) do
             (setf more-results (apply #'do-test (list c restaurant step)))
               (when (success more-results)
                  (cond ((listp (score more-results))
                         (add-to-score results (score more-results)))
                        (t
                           (add-score results (score more-results))))
               ;   (format t "~% bindings of :sup: ~a" (bindings more-results))
                  (add-bindings results (bindings more-results)))))
       results))

;*******************************************************************
;*                                                                 *
;*               WORD-COND  CLASS                                  *
;* Its do-test compares the input to the words expected by this    *
;* step. (wordc)                                                   * 
;*******************************************************************
   

(defclass word-cond (step-condition)())

(defmethod do-test ((self word-cond)  restaurant step)
    :documentation "The necessary and supporting conditions contain
      sets of words instead of lambdas. Look for each word wd from
      the necessary and supporting conditions in the input set. 
      (The words in the input set have already been looked up in the
      dictionary.) Look for the words in order first, but if they are
      not in the right order in the input then look to see if they are
      there but in the wrong order. Score 2 points for every necessary word 
      found and 1 for every supporting word found. 1 point is scored for words
      found out of order.
        Results is a 4-item object. The first item is whether or not do-test 
       succeeded, the second is the score from the test, and the third is a 
       list of any bindings that got made. The fourth item is a list of the i
       nput, with the recognized parts removed. Do-test returns these results 
        when it exits.
      NOTE: The score is a numerical list of the sentence positions 
      (first word, second  word, etc.) of the words that matched."

    (let*  ((input (input (waiter restaurant)))
            (copy input)
            (found nil)
            (count 0)
            (return-value (make-instance 'results :success t
                                            :remaining-input (copy-list input)
                                                  :score '())))
   ;      (setf copy input)         
         (when debug2p  
	  (format t "~% xxxx results going in to do-test are: ~a" return-value)
	    (format t "~%      copy of the input is: ~a" copy)
	    (format t "~%      necessary-conditions: ~a" 
            (necessary-conditions self)))
  
         (loop for wd in (necessary-conditions self) do

            (setf found (found-word wd copy count))
            (when debug2p (format t "~%    results of found-word: ~a" found))
            (cond (found 
                     (setf copy (first found))
                     (when debug2p (format t 
                           "~%   found necessary-condition ~a in do-test" wd))
                     (setf count (third found) )
                     (add-to-score return-value count)
                     (incf count)
                     (add-bindings return-value (second found))
                     (when debug2p 
                      (format t "~%looking for ~a in remaining-words: "
                               (second(second found))))
                     (use-word return-value (second (second found)))
                    (when debug2p 
                 (format t "~%return-value after changing remaining-words: ~a"
                               return-value))
                     )

                           (t
                              (setf (success return-value) nil)
                              (when debug2p 
                        (format t "~%   Couldn't find word: ~a  in input" wd ))
                              )
;;;                   ))
                  )
            )
       ;  (setf copy input)
         (loop for wd in (supporting-conditions self) do
            (setf found (found-word wd copy count))
            (cond (found
                     (setf copy (car found))
                     (when debug2p 
			   (format t "~%   found supporting-condition ~a" wd))
                     (setf count  (third found) )
                     (add-to-score return-value count)
                     (incf count)
                     (add-bindings return-value (second found))
                     (use-word return-value (second(second found))))
                  )
            )
        (when debug2p (format t "~%   returning return-value from do-test: ~a" 
                              return-value))
       (setf (remaining-input return-value) copy)
       (when debugp (format t "~%Return value: ~a" return-value))
       (setf (score return-value) (reverse (score return-value)))
      return-value))

(defun found-word (wd input count)
      (when input
        (let ((results (fw wd input)))
         (if results 
            (append results (list count))
            (found-word wd (cdr input) (incf count)))
         )))

(defun fw (wd input)
  :documentation "Assume the input looks like (<DO> <WAITER> <HAVE> <POP> 
   <QUESTION>).  The input word-items have objects associated with them so we 
    can look up their related object-class in the object hierarchy; the 
    information is not duplicated in the dictionary. The word-items should 
    also have a copy of the particular word that
    appeared in the student's input. 
    Wd is a word-name or a class-name paired with a variable to bind the input
     object to.
     Found-word returns a nil (if the word was not found) or a list of the
     form:
                   (remaining-input bindings).
     FW succeeds if the FIRST word in the input matches wd."


  (when debug2p (format t "~%In found-word with: ~a  and input: ~a" wd input))
  ;;; (defun unparsed-item (x) (listp x))

  (let ((word)
        (found '())
        (input-word (car input))
        (binding '()))
     (cond ((listp wd)
              (setf word (car wd))
              (setf binding (cdr wd)))
           (t (setf word wd)))
     (cond ((null input) (setf found nil))
           ((unparsed-item input-word)
 ;what we want to do here is to see if the word we are looking for (wd) is also
 ;unparsed. (sometimes it happens, if the word doesn't get into the dictionary,
             ;for example.) Then do a comparison.
             (cond ((and (unparsed-item word)
                         (lines-up word input))
                    (setf found (list (nth-cdr (lines-up wd input)) nil))
                    (when debug2p 
			  (format t "~%inside unparsed-item with ~a" wd)) )
;;;                  (t (when debug2p (format t "~%   input word is unparsed: 
;;;                     skipping it."))
;;;                     (setf found (found-word wd (rest input)))
;;;                     (when found (setf found (cons (cons input-word 
;;;                      (car found)) (cdr found)))))
                    ))
           ((or (and (typep input-word 'noun)
                   (kind-of (obj-class input-word) word))
                 ;the word we are looking for is a noun or pronoun
                 ;and the noun or pronoun matches the type specified in word
                (eq (get-class-name  input-word) word)) 
                  ;the word we are looking for is not a noun or pronoun
                  ;but the type matches the type specified in word.  
              (when debug2p 
                 (format t "~%    Found-word succeeds with ~a and ~a" word
                                   (get-class-name input-word))
                 )
               (if binding 
                    (setf binding (append binding (list input-word))))
               (setf found (list (rest input) (list binding)))
               (when debug2p 
		     (format t "~%    Results of found-word: ~a" found)))
;;;            (t (setf found (found-word wd (rest input)))
;;;               (when found (setf found (cons (cons input-word (car found)) 
;;;                (cdr found)))))
           )
     found))


;*********************************************************
;*               OR-COND CLASS                           *
;*               (orc)                                   *
;*********************************************************


(defclass or-cond (step-condition)())

(defmethod do-test ((self or-cond) restaurant step)
  :documentation "the or-step accepts the first condition that succeeds.
      The or-step does not use its supporting-conditions. To
      'or' a set of supporting-conditions, use a step-condition with an
      or-step as its substep. The score of an or is the score of the first
      succeeding test."
  (let ((return-value '())
        (done'())
        (remaining (necessary-conditions self))
        (current-cond))
    (loop until done do
          (setf current-cond (car remaining))
          (setf remaining (cdr remaining))
       (cond (current-cond
               (setf return-value (apply #'do-test 
					 (list current-cond restaurant step)))
               (when  (success return-value)  (setf done 't)))
             (t (setf done 't))))
    return-value))


;******************************************************
;*                                                    *
;*               AND-COND CLASS                       *
;*               (andc)                               *
;******************************************************

(defclass and-cond (step-condition)()
  (:documentation "The and-cond is the same as the step-condition, so 
     the methods don't need to be redefined. (In fact, the class 
     definition is a bit redundant, but this way and and or steps are
     symmetrical."))


;*****************************************************
;*                                                   *
;*              NOT-COND CLASS                       *
;*              (notc)                               *
;*****************************************************

(defclass not-cond (step-condition)())

(defmethod do-test ((self not-cond) restaurant step)
;  (when debugp (format t "~%    inside a not-cond"))
  (let ((return-value (make-instance 'results))
        (argument (apply #'do-test (list (necessary-conditions self) 
					 restaurant step))))
     (when (not (success argument))
;             (when debugp (format t "~%    not-step succeeded"))
             (setf (success return-value) t)
             (if (listp (score argument))
                (add-to-score return-value 1)
                (add-score return-value 1)))
     return-value))
            

;****************************************************
;*                                                  *
;*              MAX-COND CLASS                      *
;*              (maxc)                              *
;****************************************************
  
(defclass max-cond (step-condition)())

(defmethod do-test ((self max-cond) restaurant step)
  :documentation "the max-step returns the necessary-condition with
    the highest score, if there are any that succeed. Like the or-step,
    max-step is intended to be used as a substep of a step-condition."

;  (when debugp (format t "~%  Inside a max-cond"))
  (let ((return-value '())
        (best-so-far (make-instance 'results :score 0)))
    (loop  for wd in (necessary-conditions self) do 
     (setf return-value (apply #'do-test (list wd restaurant step)))
             (when (success return-value)
                   (when (greater-than (score return-value) 
				       (score best-so-far))
                        (setf best-so-far return-value))))
    best-so-far))



;****************************************************
;*                                                  *
;*       SPECIAL-SCORE-COND CLASS                   *
;*         (spcorec)                                *
;*  This is mostly for default steps.               *
;****************************************************


(defclass special-score-cond (step-condition)
  ((default-score
    :initarg :default-score
    :initform 0.5
    :accessor default-score))
  (:documentation "The special-score-step evaluates to true if it has
     no necessary-conditions; otherwise it evaluates its necessary-
     conditions to see if it applies. Its score is equal to the default-score
     regardless of the score of its necessary-conditions. This step is
     intended to be used for things like listening to the student and 
     error steps, which should only be done if there isn't some other
     step that applies. "))

(defmethod do-test ((self special-score-cond) restaurant step)
  (let ((return-value (make-instance 'results 
                                     :score (default-score self)
                                     :success t)))
        (when (necessary-conditions self)
                 (setf return-value (apply #'do-test (list
                                                        (necessary-conditions 
							 self) 
                                                        restaurant
                                                        step)))
                 (setf (score return-value) (default-score self)))
        return-value))
              


;*****************************************************
;*                                                   *
;*            HISTORY-COND CLASS                     *
;*             (historyc)                            *
;*  Checks if something is on the waiter's history   *
;* list.                                             *
;*****************************************************


(defclass history-cond (step-condition)())

(defmethod do-test ((self history-cond) restaurant step)
  :documentation "There should be one condition in the necessary conditions of
     the history test. If that condition is met, then the step is successful: 
score 1."
;  (when debugp (format t "~%    inside a history-test"))
  (let ((current-history (history (waiter restaurant)))
        (return-value (make-instance 'results)))
    (when (my-subtest self current-history (necessary-conditions self))
;                 (when debugp (format t "~%     found ~a in the history" 
;                                     (necessary-conditions self)))
                 (setf (success return-value) t)
                 (setf (score return-value) 1))
    return-value))

(defmethod my-subtest ((self history-cond) current-history item-to-find)
  :documentation "The subtest compares each item in the waiter's history-list
     to the item is should be looking for. When a match is found, it returns
     true. If no match is found, it returns nil. 

     NOTE: The item-to-find should be the name of a step, i.e. a string."

  (cond ((null current-history) nil)
        ((null (car current-history)) (my-subtest self (cdr current-history) 
						  item-to-find))
        ((string-equal (name (car current-history)) item-to-find) t)
        (t (my-subtest self (cdr current-history) item-to-find))))


;******************************************************
;*                                                    *
;*              LAMBDA-COND CLASS                     *
;*              (lambdac)                             *
;* Tests an arbitrary function. The function must     *
;* take the arguments input and restaurant and must   *
;* return a result-object)                            *
;******************************************************


(defclass lambda-cond (step-condition)()
  (:documentation "The lambda-cond tests an arbitrary function. It can have
     both necessary and supporting conditions, unlike or, max, history and
     special-score. The functions (one in necessary-conditions and an optional 
     one in supporting-conditions) must take args (i.e. :restaurant restaurant)
     as its keyword arguments and must return a result-object regardless of the
     success or failure of the function."))

(defmethod do-test ((self lambda-cond) restaurant step)
  (let ((return-value '())
        (supporting-evidence '()))
        (setf return-value (funcall (necessary-conditions self)  
				    restaurant step))
        (when (and (success return-value) (supporting-conditions self))
           (setf supporting-evidence (funcall (supporting-conditions self) 
					      restaurant step))
           (when (success supporting-evidence)
              (add-score return-value (score supporting-evidence))
              (add-bindings return-value (bindings supporting-evidence))))
     return-value))

;-------------------------------------------------------------------------
;  Macros to make step-conditions easier to use
;-------------------------------------------------------------------------

(defun stepc (nec necs &optional sup supporting)
    (cond ((eq nec ':nec) 
           (if (and sup (eq sup ':sup)) 
               (make-instance 'step-condition
                              :necessary-conditions necs
                              :supporting-conditions supporting)
             (make-instance 'step-condition
                            :necessary-conditions necs)))
          ((eq nec ':sup)
             (if sup
               (make-instance 'step-condition
                          :necessary-conditions supporting
                          :supporting-conditions necs)
               (format t "stepc needs to have supporting conditions")))
          (t (format t "format for stepc: (stepc :nec necessary-conditions 
                                               :sup supporting-conditions)"))))

(defun wordc (nec necs &optional sup  supporting) 
    (cond ((eq nec ':nec) 
           (if (and sup (eq sup ':sup)) 
               (make-instance 'word-cond
                              :necessary-conditions necs
                              :supporting-conditions supporting)
             (make-instance 'word-cond
                            :necessary-conditions necs)))
          ((eq nec ':sup)
             (if sup
               (make-instance 'word-cond
                          :necessary-conditions supporting
                          :supporting-conditions necs)
               (format t "wordsc needs to have supporting conditions")))
          (t (format t "format for wordsc: (wordsc :nec necessary-conditions 
                                                 :sup supporting-conditions)")
             (format t "~%(you wrote ~a ~a ~a ~a)" nec necs sup supporting))))
    
(defun orc (conditions)
  (make-instance 'or-cond
     :necessary-conditions conditions))

(defun andc (nec necs &optional sup supporting)
    (cond ((eq nec ':nec) 
           (if (and sup (eq sup ':sup)) 
               (make-instance 'and-cond
                              :necessary-conditions necs
                              :supporting-conditions supporting)
             (make-instance 'and-cond
                            :necessary-conditions necs)))
          ((eq nec ':sup)
             (if sup
               (make-instance 'and-cond
                          :necessary-conditions supporting
                          :supporting-conditions necs)
               (format t "andc needs to have supporting conditions")))
          (t (format t "format for andsc: (andsc :nec necessary-conditions 
                                               :sup supporting-conditions)"))))

(defun notc ( condition)
  (make-instance 'not-cond
     :necessary-conditions condition))

(defun maxc ( condition)
  (make-instance 'max-cond
     :necessary-conditions condition))

(defun spscorec (&optional sc score nec necs )
; (when debugp 
 ; (format t "~%conditions in spscorec: ~a ~a ~a ~a" sc score nec necs))
    (cond ((eq sc ':sc) 
           (if (and nec (eq nec ':nec)) 
               (make-instance 'special-score-cond
                              :necessary-conditions necs
                              :default-score score)
             (make-instance 'special-score-cond
                            :default-score score)))
          ((eq sc ':nec)
             (if (and nec (eq nec ':sc))
               (make-instance 'special-score-cond
                          :necessary-conditions score
                          :default-score necs)
                (make-instance 'special-score-cond
                          :necessary-conditions score)))
          (t (make-instance 'special-score-cond))))

(defun historyc (condition)
  (make-instance 'history-cond
     :necessary-conditions condition))

(defun lambdac (condition)
  (make-instance 'lambda-cond
     :necessary-conditions condition))














