;######################################################################
;#                                                                    #
;#                      GRAPH.LSP                                     #
;#                                                                    #
;#  Contents:                                                         #
;#      Functions:  find-optimal-step-cover (candidate-steps)         #
;#                  graph-test                                        #
;#      Classes:  graph, node, link                                   #
;#                                                                    #
;######################################################################

;----------------------------------------------------------------------------
;   find-optimal-step-cover
;---------------------------------------------------------------------------

                          
(defun find-optimal-step-cover (candidate-steps)
   "We want to find the set of steps that 'optimally' covers the input.
    We want to get a set of steps that doesn't over lap but has the highest
    possible weighting score. We can view this as a find the longest path
    problem. We want to find the longest path from the beginning of a sentence
    that goes through the biggest (weightiest) or longest steps to the
    end of the sentence. We can convert this into a shortest path problem
    by simply negating all of the weights. 

    We first construct a graph from the candidate steps. We create two
    artificial nodes called /start/ and /end/. Then we create a node
    in the graph for every candidate step. Next we join /start/ to every
    candidate node with a link having zero length so that a path starting
    with any step is equally likely. We then add paths from every candidate
    node to /end/ with weight equal to the weight of the step. This ensures
    that any node can end the valid step sequence, but the cost of the step
    is taken into account. Finally, we add links between the candidate nodes
    exactly where one node strictly precedes the second node. To strictly 
    precede a node, the last word index of the first node must be before
    the first word index of the second node. We then find the shortest
    path, copy the contents of the graph nodes back out into a list and
    voila: the optimal set of steps."
    
   (Let* ((graph (make-instance 'graph))
          (start (addItem graph nil "Start"))
          (end   (addItem graph nil "End"))
          (current nil)
          (shortest-path nil)
          (steps nil))
      
      ;;; First, construct null path from start to every step and
      ;;; from every step to the end node.
         (loop for step in candidate-steps do
            (setf current (addItem graph step (name step)))
            (addLink start current 0)
    ;        (format t "~%Constructing ~a" current)
            (addLink current end (- 1000 (step-weight (contents current)))))
      
      ;;; Now construct all valid paths
      
         (loop for n1 in (remove start (remove end (nodes graph))) do
            (loop for n2 in (remove start (remove end (nodes graph))) do
               (when (strictly-precedes (contents n1) (contents n2))
     ;             (format t "~%Linking preceding step ~a" n1)
                  (addLink n1 n2 (- 1000 (step-weight (contents n1))) ))))
      
      ;   (format t "~%candidate set ~a ~%" candidate-steps)
     ;    (dump graph)
      
      ;;; Find Shortest Path and return it
         (setf shortest-path (second (shortest-path graph start end)))
         (setf shortest-path (delete start shortest-path))
         (setf shortest-path (delete end shortest-path))
         (setf steps (loop for n in shortest-path collect (contents n)))
      
    ;  (format t "~%Shortest path: ~a" shortest-path)
    ;  (format t "~%Steps (inside graph.lsp): ~a" steps)
      steps
   )
)
            
;----------------------------------------------------------------------------
;   Graph
;----------------------------------------------------------------------------

(defclass graph ()
    ((nodes :initform '() :accessor nodes :initarg :nodes)))

(defmethod addNode ((self graph) node)
   (push node (nodes self))
)

(defmethod addItem ((self graph) item name)
   "Encapsulates an item in an abstract node, and returns a pointer to node"
   (let ((new-node (make-instance 'node :contents item :name name)))
      (addNode self new-node)
      new-node))
                     
                     
(defmethod addNode ((self graph) node)
   "Used to add an abstract graph node to the graph"
   (push node (nodes self))
)

(defmethod print-object ((self graph) s)
   (format s "<Graph:~%")
   (loop for n in (nodes self) do
      (format s "~%")
      (print-object n s))
   (format s ">"))

(defmethod dump ((self graph) )
   (format t "<Graph:~%")
   (loop for n in (nodes self) do
      (format t "~%")
      (dump n))
   (format t ">"))

(defmethod shortest-path ((self graph) start-node end-node)
   ;;;finds the shortest path through the graph from start-node to end-node,
   ;;;using Dijkstra's algorithm (or some form thereof)
  
   (let ((S '())
         (U (delete start-node (nodes self)))
         (min nil)
         (minvert nil)
         (distance 0)
         (shortest-path '())
         (focus nil))
         
         (loop for n in U do 
            (setf (estimate n) 'nil))
         
         (setf (minDist start-node) 0)
         (setf focus start-node)
      
         (loop until (member end-node S) do
            (setf min nil)
            ;;(format t "~%Begin iteration (U:~a)" U)
            (loop for vertex in U do
               (let ((distance (distanceTo focus vertex)))
 ;;(format t "~%   focus:~a   minDist focus:~a    vertex:~a   
;;  minDist vertex: ~a  dist focus to vertex: ~a"
                  ;;   focus (minDist focus) vertex (minDist vertex) distance)
                  (cond ((and 
                           distance
                           (estimate vertex)
                           ( < (+ (minDist focus) distance) (estimate vertex)))
                             (setf (minNeighbour vertex) focus)
                             (setf (estimate vertex) 
				   (+ (minDist focus) distance)))
                        ((and distance (null (estimate vertex)))
                           (setf (minNeighbour vertex) focus)
                           (setf (estimate vertex) 
				 (+ (minDist focus) distance))))
                  
                  (cond ((and min (estimate vertex) (< (estimate vertex) min))
                         (setf min (estimate vertex))
                         (setf minVert vertex))
                        ((and (null min) (estimate vertex))
                         (setf min (estimate vertex))
                         (setf minVert vertex)))
            ))
            
            
            ;; Update the min distance of the vertex with minimal distance
                        
            (setf (minDist minVert) (estimate minVert))
            (push minVert S)
            (setf U (delete minVert U))
            (setf focus minVert)
            ;;(format t "~%~%")
        )
      
        ;; Reverse construct the shortest path
        (loop with n = end-node until (eq n start-node) do
           (push n shortest-path)
           (setf n (minNeighbour n)))
        (push start-node shortest-path)
        (setf distance (minDist end-node))
        (list distance shortest-path)
    )                 
)



(defun graph-test ()
   (let* ((g (make-instance 'graph))
          (start (addItem g "A" "start"))
          (end (addItem g "A" "end"))

                    (kinds-of (addItem g "A" "kinds-of"))
          (do-you-have (addItem g "B" "do-you-have"))
          (get-item (addItem g "C" "get-item"))
          (describe-item (addItem g "C" "describe-item"))
          (list-menu (addItem g "f" "list-menu"))

                   ; (n1 (addItem g "A" "n1"))
         ; (n2 (additem g "B" "n2"))
         ; (n3 (additem g "C" "n3"))
         ; (n4 (additem g "D" "n4"))
         ; (n5 (additem g "E" "n5"))
          (result nil)
          (items nil))
      
;      (addLink n1 n2 10)
;      (addLink n1 n4 30)
;      (addLink n1 n5 100)
;      (addLink n2 n3 50)
;      (addLink n3 n5 10)
;      (addLink n4 n5 60)
;      (addLink n4 n3 20)

       (addLink start kinds-of 0)
       (addLink start get-item 0)
       (addLink start describe-item 0)
       (addLink start do-you-have 0)
       (addLink start list-menu 0)
      
       (addLink  kinds-of end -4)
       (addLink  get-item end -1)
       (addLink  describe-item end -2)
       (addLink  do-you-have end -2)
       (addLink  list-menu end -2)

       (addLink  get-item do-you-have -1)
       (addLink  describe-item do-you-have -2)

      (format t "~%Created a new Graph~%")
      (dump g )
      (format t "~%~%")
      (setf result (shortest-path g start end))
      (setf items (loop for item in (second result) collect (contents item))) 
     
      (format t"~%Shortest distance: ~a and items: ~a" result items)
   )
)

;----------------------------------------------------------------------------
;   Node
;----------------------------------------------------------------------------

(defclass node()
    ((links :accessor links :initform '() :initarg :links)
     (contents :accessor contents :initform nil :initarg :contents)
     (minDist :accessor minDist :initform nil :initarg :minDist)
     (estimate :accessor estimate :initform nil :initarg :estimate)
     (minNeighbour :accessor minNeighbour :initform nil :initarg :minNeighbour)
     (name :accessor name :initform "" :initarg :name))
   )

(defmethod addLink ((self node) node weight)
   (push (make-instance 'link :node node :weight weight) (links self))
)

(defmethod distanceTo ((self node) (other node))
   "Returns distance if self is connected directionally to other,
    otherwise returns nil"
   (loop for link in (links self) 
      thereis 
        (if (eq (node link) other) (weight link))))
           
           
(defmethod print-object ((self node) s)
      (format s "<~a>" (name self)))

(defmethod dump ((self node))
      (format t "<~a:" (name self))
      (loop for n in (links self) do
          (dump n))
      (format t ">"))

;----------------------------------------------------------------------------
;   Link
;----------------------------------------------------------------------------

(defclass link ()
   ((node :initform '() :accessor node :initarg :node)
    (weight :initform 0 :accessor weight :initarg :weight)))

(defmethod print-object ((self link) s)
   (format s "<Link: -(~a)-> ~a>"  (weight self) (name (node self))))


(defmethod dump ((self link) )
   (format t "--(~a)--> ~a "  (weight self) (name (node self))))














