;##########################################################################
;#                                                                        #
;#                         UTIL.LSP                                       #
;#                                                                        #
;#  CONTENTS:                                                             #
;#       Functions:  random-pick (list), flatten (list),                  #
;#                   append-char (string char), greater-than (a,b)        #
;#                   string-to-list (string), list-to-string (list)       #
;#                   lines-up (a b), nth-cdr (index list)                 #
;#                   kind-of ((subclass symbol)( superclass symbol))      #
;#                   wait (seconds), get-most-frequent (list)             #
;#                   prev-step-matches (restaurant list-of-steps new-score#
;#                   step-appears (restaurant step num-times new-score    #
;#                   unparsed-item (X), top (stack),                      #
;#                   dictionary-size (dict)                               #
;#      Parameters:  yes, no                                              #
;#      Classes:    pair                                                  #
;#                                                                        #
;#      Comment:    This file contains a number of usefull functions that #
;#                  are used in parsing and in the my-conditions of the   #
;#                  order scenarios.                                      #
;#                                                                        #
;##########################################################################


(defun random-pick (list)
;; Function returns the nth randomly generated element of a list.
;; This is used to generate random output from the waiter.
   (nth (random (length list)) list))

(defun greater-than (a b)
   (cond ((and (listp a)
               (listp b))
          (> (length a) (length b)))
         ((listp a)
          (> (length a) b))
         ((listp b)
          (> a (length b)))
          (t (> a b))))

(defun flatten (list)
;;Flattens a list of list into one big list.
    (let ((result '()))
      (cond ((null list) nil)
            ((not (listp (car list) )) 
             (cons (car list) (flatten (cdr list))))
            (t (append (flatten (car list)) (flatten (cdr list)))))))
            

(defun append-char (string char)
;; Used in function "list-to-string".  Appends char to the end of the list 
;;string.
   (format nil "~a~a" string char))

(defun string-to-list (string)
;; Turns a string into a list.  This is used in the function "walk-string" in
;; the file "atn.lsp".  "walk-string" in parsing to turn the user input into
;; a list of words.
   (let ((list '()))
      (loop for i downfrom (- (length string) 1) to 0
           do (push (aref string i) list))
      list))

(defun list-to-string (list)
;; Turns a list into a string.  This is in the file "dictiona.lsp" to set 
;; the reference names of items.
   (let ((string ""))
      (dotimes (i (length list))
         (cond ((listp (nth i list) )
                   (setf string (append-char string (list-to-string 
						     (nth i list))))
                   (setf string (append-char string " "))  )
               (t (setf string (append-char string (nth i list))))))
      (string-trim '(#\space) string)))


(defun lines-up (a b)
;; This function is also used in parsing to see if special form of words
;; are present in the input.
   :documentation "(lines-up a b) returns the length of a (i.e. the amount of
    a that matched b)  if a is a sublist of b and a starts at the start of b."
  (let ((found t))
;;;  (when debugr (format t "~%inside lines-up with ~a and ~a" a b))
   (loop for i from 0 below (length a) do 
         (unless (equal (nth i a) (nth i b))
                 (setf found nil)))
   (if found (setf found (length a)) found)))


(defun nth-cdr (index list)
   :documentation "return the tail of the list. The 'tail' is what follows
   the list after and including the indexth item in the list. For example,
   (nth-cdr 0 '(a b c)) will return '(a b c);
   (nth-cdr 1 '(a b c)) will return '(b c), and so on. 
   (note: (nth-cdr 5 '(a b c)) will return '())"
;;;(when debugr (format t "~%inside nth-cdr with ~a and ~a" index list))
  (if (<= index 0)
      list
      (nth-cdr (- index 1) (cdr list))))

;;;(defmethod kind-of ((subclass symbol) (superclass symbol))
   ;;;USE THIS VERSION IN ALLEGRO COMMON LISP
;;;    :documentation "returns true if the first argument is a subclass of the
 ;;;        second"
;;;      (loop with class = (find-class subclass)
;;;	    until (or (eq (class-name class) 'standard-object)
;;;		      (eq (class-name class) t))
;;;	    do 
;;;	    (when debug2p (format t "~%(In Kind-of) Class:~a ~a" 
;;;(class-name class) superclass))
;;;	    (when  (eq  superclass (class-name class))
;;;	      (return t)) 
;;;	    (setf class (first (class-direct-superclasses class)))))

 (defmethod kind-of ((subclass symbol) (superclass symbol))
     ;;;USE THIS VERSION IN 
     :documentation "returns true if the first argument is a subclass of the 
      second"
       (loop with class = (find-class subclass)
 	    until (or (eq (class-name class) 'standard-object)
 		      (eq (class-name class) t))
 	    do 
 	    (when debug2p (format t "~%(In Kind-of) Class:~a ~a" (class-name 
                            class) superclass))
 	    (when  (eq  superclass (class-name class))
 	      (return t)) 
 	    (setf class (first (clos:class-direct-superclasses class)))))

(defun wait (seconds) 
;; Makes the computer wait the specified number of seconds.  This is used in 
;; the serve scenario.
  (let ((start-time (+ (get-universal-time) seconds)))
    (loop until (> (get-universal-time) start-time) do
        ;  (format t "~% ~a" (listen *terminal-io*))
       (+ 1 1)
       )))

(defun get-most-frequent (list)
  :documentation "return the most frequent element of the list. If there is a
    tie, the first of the tied elements that occurs in the list is returned.
    Nil is returned if the list is empty.
    Used by the Waiter::get-best-scenario method. "
  (if list
      (let ((unique-elements '())
            (scores '())
            (temp '())
            (index 0))
        (loop for i in list do
              (setf temp (member i unique-elements))
              (cond (temp (setf index (- (length unique-elements) (length temp
									  )))
                          (incf (nth index scores) ))
                    (t (push i unique-elements) 
                       (push 1 scores))))
        (setf scores (reverse scores))
        (setf unique-elements (reverse unique-elements))
        (setf index (-  (length scores) (length (member (apply 'max scores) 
							scores))))
        (nth index unique-elements))))

(defun prev-step-matches (restaurant list-of-steps new-score)
;; If the name of the last step on the waiter's history list matches one of 
;;the step names in list-of-step an instance of the result class is created 
;;and returned with success set to t, and the score set to new-score.  The
;; function is used frequently in the my-conditions of steps in the order
;; scenarios.

  ; (format t "~%Inside prev-step-matches")
   (let* ((results (make-instance 'results :success nil))
          (prev (last-step (waiter restaurant) ))
         (a-step (if prev (name prev) nil)))
   ;  (format t "~%Inside let-statement of prev-step-matches")
      (when a-step
         (loop for a in list-of-steps until (success results) do
   ;        (format t "~%comparing ~a and ~a" a-step a)
            (when (string-equal a-step a)
               (setf (success results) t)
               (setf (score results) new-score))))
   ;   (format t "~%Exitting prev-step-matches with: ~a" results)
      results))

(defun step-appears (restaurant step num-times new-score)
;; If the step appears more than num-times in the waiter's history list,
;; an instance of the results class is returned with success set to t, and
;; the score set to new-score.  This used used in the my-conditions of the
;;"ask-drink-order", "ask-food-order" and "ask-dessert-order".  It is used
;; to make sure that the steps can be repeated, but not over and over.

;; (format t "~%Inside step-appears with step ~a" step )
   (let ((results (make-instance 'results :success nil))
        (count 0))
    (loop for a-step in (history (waiter restaurant)) do
      (if (equal step (name a-step))
         (setf count (+ 1 count))))
;;(format t "~%Count: ~a" count)
    (cond ((>= count num-times)
        (setf (success results) t)
        (setf (score results) new-score)))
;;(format t "~%Exitting step-appears with : ~a" results)
results)) 


   
   
(defun unparsed-item (X)   (listp X))

(defparameter yes t)
(defparameter no nil)
  
(defun top (stack)
         (car stack))


;;The pair class is used in the focus-type functions in the file "general.sce".

(defclass pair ()
    ((head 
        :accessor head
        :initarg :head
        :initform nil)
    (tail 
       :accessor tail
       :initarg :tail
       :initform nil)))

(defmethod as-list ((self pair))
   (list (head self) (tail self)))

(defmethod print-object ((self pair) stream)
   (format stream "<Pair: ~a ~a>" 
      (head self) (tail self)))

(defun dictionary-size (dict)
;; Returns the size of a dictionary including the special-forms and 
;; special-phrases
        (let ((sum 0))
        (loop for i in (contents dict) do
            (incf sum)
            (setf sum (+ sum (length (special-forms i))))
            (setf sum (+ sum (length (special-phrases i )))))
         sum))











